package POE::Session::Multiplex;

use 5.008008;
use strict;
use warnings;

use Carp;
use POE;
use POE::Session;

use Scalar::Util qw( blessed );

require Exporter;

our @ISA = qw(Exporter POE::Session);
our %EXPORT_TAGS = ( 'all' => [ qw( ev evs rsvp evo evos ) ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw( ev evs rsvp );

our $VERSION = '0.0100';
our $CURRENTOBJ;

#######################################
BEGIN {
    *SE_NAMESPACE = \&POE::Session::SE_NAMESPACE;
    *SE_OPTIONS   = \&POE::Session::SE_OPTIONS;
    *SE_STATES    = \&POE::Session::SE_STATES;

    *EN_SIGNAL    = \&POE::Session::EN_SIGNAL;
    *EN_DEFAULT   = \&POE::Session::EN_DEFAULT;

    *OPT_TRACE    = \&POE::Session::OPT_TRACE;
    *OPT_DEBUG    = \&POE::Session::OPT_DEBUG;
    *OPT_DEFAULT  = \&POE::Session::OPT_DEFAULT;
}

sub SE_OBJECTS () { SE_STATES+1 }
sub SE_STATERE () { SE_STATES+2 }

sub OH_OBJECT () { 0 }
sub OH_NAME   () { 1 }
sub OH_OURS   () { 2 }

############################################################################
sub _loggable
{
    my( $self ) = @_;
    $POE::Kernel::poe_kernel->_data_alias_loggable( $_[0] );
}

sub instantiate
{
    my $self = shift->SUPER::instantiate( @_ );
    if( $self->[SE_OBJECTS] ) {
        die "Definition of POE::Session changed!  $self needs to be modified.\n";
    }
    $self->[SE_OBJECTS] = {};
    return $self;
}

#######################################
sub _invoke_state
{
    my( $self, $source_session, $state, $etc, $file, $line, $fromstate ) = @_;

    $self->set_objectre unless $self->[SE_STATERE];

    if ($self->[SE_OPTIONS]->{+OPT_TRACE}) {
        POE::Kernel::_warn( $self->_loggable,
                            " -> $state (from $file at $line)\n"
                          );
    }
    
    if( $state =~ /^$self->[SE_STATERE]$/ ) {
        my $obj_name = $1;
        my $obj_state = $2;

        my( $object, $method ) = $self->handler_for( $obj_name, $obj_state, $state );
        if( $method ) {
            if ($self->[SE_OPTIONS]->{+OPT_TRACE}) {
                POE::Kernel::_warn( $self->_loggable,
                                    " OBJECT=$obj_name METHOD=$method\n"
                                  );
            }

            local $CURRENTOBJ = $obj_name;
            return $object->$method(
                    $self,                          # SESSION
                    $POE::Kernel::poe_kernel,       # KERNEL
                    $self->[SE_NAMESPACE],          # HEAP
                    $obj_state,                     # STATE
                    $source_session,                # SENDER
                    undef,                          # unused #6
                    $file,                          # CALLER_FILE
                    $line,                          # CALLER_LINE
                    $fromstate,                     # CALLER_STATE
                    @$etc,                          # ARG0, ARG1, ARG2
              );
        }
    }

    if ($self->[SE_OPTIONS]->{+OPT_TRACE}) {
                POE::Kernel::_warn( $self->_loggable,
                                " -> $state (not an multiplex state)\n"
                          );
    }
    ## Skip out if it's not our state
    return shift->SUPER::_invoke_state( @_ );
}


#######################################
sub evo ($$)
{
    my( $obj_name, $event ) = @_;
    return "$obj_name->$event";    
}

#######################################
sub ev ($)
{
    my( $event ) = @_;
    croak "Can't call ev outside of a multiplexed object" unless $CURRENTOBJ;
    return evo $CURRENTOBJ, $event;
}
#######################################
sub evs ($)
{
    my( $event ) = @_;
    croak "Can't call evs outside of a multiplexed object" unless $CURRENTOBJ;
    my $self = $poe_kernel->get_active_session;
    croak "Can't call evs outside of a multiplexed session" unless $self;
    return $self->ID, ev$event;
}

sub rsvp ($)
{
    return [ evs $_[0] ];
}

#######################################
sub evos ($$$)
{
    my( $session, $obj_name, $event ) = @_;
    $session = $session->ID if blessed $session;
    return $session, evo $obj_name, $event;
}


#######################################
sub set_objectre
{
    my( $self ) = @_;
    my $re = '(' . join( '|', map quotemeta, keys %{$self->[SE_OBJECTS]} ).
              ')->(.+)';
    $self->[SE_STATERE] = $re;
}

#######################################
sub handler_for
{
    my( $self, $obj_name, $obj_state, $state ) = @_;

    my @ret = $self->_handler_for( $obj_name, $state );
    return @ret if @ret;

    return $self->_handler_for( $obj_name, $obj_state );
}

sub _handler_for
{
    my( $self, $obj_name, $state ) = @_;
    my $handler = $self->[SE_STATES]->{$state};
    return unless $handler;
    return if 'CODE' eq ref $handler;

    my $def = $self->[SE_OBJECTS]->{$obj_name};
    return unless $def;

    # Same object?
    if( blessed $handler->[0] and $handler->[0] eq $def->[OH_OBJECT] ) {
        return $def->[OH_OBJECT], $handler->[1];
    }
    # Object of the right package?
    elsif( $def->[OH_OBJECT]->isa( $handler->[0] ) ) {
        return $def->[OH_OBJECT], $handler->[1];
    }
    return;
}

#######################################
sub object
{
    my( $self, $obj_name, @def ) = @_;
    if( @def ) {
        if( blessed $obj_name and $obj_name->can( 'name' ) ) {
            $def[0] = $obj_name;
            $obj_name = $def[0]->name;
        }
        $self->object_register( name => $obj_name, 
                                object => $def[0], 
                                states => $def[1]
                              );
    }
    else {
        $self->object_unregister( $obj_name );
    }
}

#######################################
sub object_get
{
    my( $self, $obj_name ) = @_;
    return unless exists $self->[SE_OBJECTS]->{ $obj_name };
    return $self->[SE_OBJECTS]->{ $obj_name }[OH_OBJECT];
}

#######################################
sub object_register
{
    my( $self, %def ) = @_;

    my $obj_name = $def{name};
    my $object   = $def{object};
    my $states   = $def{states} || $def{events};
    return unless $object;

    if( $self->[SE_OBJECTS]->{ $obj_name } ) {
        $self->object_unregister( $obj_name );
    }
    $self->[SE_STATERE] = '';

    my $hold;
    if( $states ) {
        # Tediously define states for this object
        $hold = $self->[SE_OBJECTS]->{ $obj_name } = [];
        my $ours = $hold->[OH_OURS] = [];
        local $CURRENTOBJ = $obj_name;
        $states = [ $states ] unless ref $states;
        if( 'HASH' eq ref $states ) {
            while( my( $event, $method ) = each %$states ) {
                push @$ours, ev$event;
                $self->_register_state( $ours->[-1], $object, $method );
            }
        }
        else {
            foreach my $event ( @$states ) {
                push @$ours, ev$event;
                $self->_register_state( $ours->[-1], $object, $event );
            }
        }
    }
    else {
        # Make sure there are some states defined for this object's class
        my $ok = 0;
        my $package = ref $object;
        foreach my $handler ( values %{ $self->[SE_STATES] } ) {
            next unless 'ARRAY' eq ref $handler;
            next if blessed $handler->[0] and $handler->[0] ne $object;
            next unless $package eq $handler->[0] or 
                        $object->isa( $handler->[0] );
            $ok = 1;
            last;
        }
        croak "No package_states defined for package $package" unless $ok;
        $hold = $self->[SE_OBJECTS]->{ $obj_name } = [];
    }
    $hold->[OH_NAME]   = $obj_name;
    $hold->[OH_OBJECT] = $object;
    return 1;
}

#######################################
sub object_unregister
{
    my( $self, $obj_name ) = @_;
    if( blessed $obj_name and $obj_name->can( 'name' ) ) {
        $obj_name = $obj_name->name;
    }
    my $def = delete $self->[SE_OBJECTS]->{ $obj_name };
    $self->[SE_STATERE] = '';
    return unless $def and $def->[OH_OURS];

    foreach my $event ( @{ $def->[OH_OURS] } ) {
        $self->_register_state( $event );
    }        
    return 1;
}

#######################################
sub package_register
{
    my( $self, $package, $states ) = @_;

    $states = [ $states ] unless ref $states;
    if( 'HASH' eq ref $states ) {
        while( my( $event, $method ) = each %$states ) {
            $self->_register_state( $event, $package, $method );
        }
    }
    else {
        foreach my $event ( @$states ) {
            $self->_register_state( $event, $package, $event );
        }
    }
    return 1;
}

1;

__END__

=head1 NAME

POE::Session::Multiplex - POE session with object multiplexing

=head1 SYNOPSIS

    use POE;
    use POE::Session::Multiplex;

    My::Component->spawn( @args );
    $poe_kernel->run;

    package My::Component;

    sub spawn {
        my( $package, @args ) = @_;
        POE::Session::Multiplex->create(
                        package_states => [
                                $package => [ qw( _start ) ]
                                'My::Component::Object' => 
                                        [qw( fetch decode back )] 
                            ]
                        args => [ @args ]
                    );
    }

    ##### To add an object to the session:
    my $obj = My::Component::Object->new();
    $_[SESSION]->object( $name, $obj );
        # or
    $_[SESSION]->object_register( name => $name, object => $obj );

    ##### To remove an object
    $_[SESSION]->object_unregister( $name );


    ##### To get a state name in the current object:
    my $state = ev "state";

    ##### To build a session/state tuple for the current object
    my( $session, $state ) = evs "state";
    
    # this tuple would be useful for getting a response from another session
    $poe_kernel->post( $session=>'fetch', $params, rsvp "fetch_back" );
    
    # and in fetch, you could reply with simply:
    $poe_kernel->post( @$rsvp, $reply );

    ##### Posting to a specific object from inside the session
    $poe_kernel->post( evo( $name, $state ), @args );

    ##### Posting to a specific object from outside the session
    use POE::Session::Multiplex qw( evos );
    $poe_kernel->post( evos( $session, $name, $state ), @args );

=head1 DESCRIPTION


=head1 METHODS

=head2 create

=head2 object_register

=head2 object_get

=head2 object_unregister

=head2 package_register

=head1 EXPORT

=head2 ev

=head2 evo

=head2 evs

=head2 rsvp

=head2 evos

=head1 SEE ALSO

L<POE::Session>.

=head1 AUTHOR

Philip Gwyn, E<lt>gwyn-at-cpan.orgE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009 by Philip Gwyn

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.8 or,
at your option, any later version of Perl 5 you may have available.


=cut
